/*
 * Simple binary tree implementation for variable bit-length keys
 * The empty binary tree is the NULL pointer.
 */
/* David Leonard, 2003. Public domain. */

#include <stdio.h>
#include <stdlib.h>
#include <err.h>

#include "btree.h"

static const char rcs[] = "$Id: btree.c,v 1.1.1.1 2003/07/27 01:43:48 d Exp $";

struct btree {
	void	*value;
	struct btree *child[2];
	int	len;			/* length of name in bits */
	unsigned char name[1];
};

/*
 * Allocate a new binary tree node
 */
static struct btree *
btree_alloc(name, len, value)
	unsigned char *name;
	int len;	/* length of name in bits */
	void *value;
{
	struct btree *b;

	b = (struct btree *)malloc(sizeof (struct btree) + (len+7)/8 - 1);
	if (b == NULL)
		errx(1, "malloc");
	b->value = value;
	b->child[0] = NULL;
	b->child[1] = NULL;
	b->len = len;
	if (len)
	    memcpy(b->name, name, (len+7)/8);
	return b;
}

/*
 * Return true if bits 0..hi-1 of n1
 * are the same as n2. (assumes bits 0..lo-1 are already identical)
 */
static int
match(n1, n2, lo, hi)
	unsigned char *n1, *n2;
	int lo, hi;
{
	static unsigned char 
	    himask[9] = { 0x00, 0x01, 0x03, 0x07, 0x0f, 
				0x1f, 0x3f, 0x7f, 0xff };

	int lobyte = lo/8, hibyte = hi/8;
	unsigned char mask = himask[hi%8];
	int pos;

	for (pos = lobyte; pos < hibyte; pos++)
		if (n1[pos] != n2[pos])
			return 0;
	return !mask || ((n1[pos] & mask) == (n2[pos] & mask));
}

/*
 * Return the longest common name of n1 and n2,
 * assuming the are identical in the first 'lo' bits.
 * and that the shorter one is 'hi' bits long.
 */
static int
lcp(n1, n2, lo, hi)
	unsigned char *n1, *n2;
{
	int pos;

	for (pos = lo; pos < hi; pos++) {
		int byte = pos/8;
		int bit = 1 << (pos%8);
		if ((n1[byte] & bit) != (n2[byte] & bit))
			break;
	}
	return pos;
}

/*
 * Free all the binary tree nodes
 */
void
btree_free(b)
	struct btree *b;
{
	if (b) {
		btree_free(b->child[0]);
		btree_free(b->child[1]);
		free(b);
	}
}

/*
 * Return the value associated with the name, or NULL
 */
void *
btree_find(b, name, len)
	struct btree *b;
	unsigned char *name;
	int len;
{
	int prevlen = 0;

	while (b) {
		if (len < b->len || 
		    !match(name, b->name, prevlen, b->len))
			return NULL;
		if (len == b->len)
			return b->value;
		prevlen = b->len;
		if (name[prevlen/8] & (1<<(prevlen%8)))
			b = b->child[1];
		else
			b = b->child[0];
	}
	return NULL;
}

/*
 * Apply the given function to each node in the tree, in order
 */
void
btree_apply(b, fn, d)
	struct btree *b;
	void (*fn)(unsigned char *, int, void *, void *);
	void *d;
{
	while (b) {
		if (b->value)
			(*fn)(b->name, b->len, b->value, d);
		btree_apply(b->child[0], fn, d);
		b = b->child[1];
	}
}

/*
 * Insert a new node into the tree (or replace an old one).
 * Returns old value for the name (or NULL)
 */
void *
btree_insert(bp, name, len, value)
	struct btree **bp;
	unsigned char *name;
	int len;
	void *value;
{
	int prevlen = 0, lclen;
	struct btree *b, *b1, *b2;
	void *oldvalue;

	for (;;) {
		b = *bp;
		if (!b) {
			if (value)
				*bp = btree_alloc(name, len, value);
			return NULL;
		}
		if (len < b->len || 
		    !match(name, b->name, prevlen, b->len))
		{
			if (value == NULL) 
				return NULL;
			lclen = len < b->len ? len : b->len;
			lclen = lcp(name, b->name, prevlen, lclen);
			/* the two names differ at the lclen bit */
			if (len == lclen) {
				b1 = btree_alloc(name, lclen, value);
				b2 = NULL;
			} else {
				b1 = btree_alloc(name, lclen, NULL);
				b2 = btree_alloc(name, len, value);
			}
			*bp = b1;
			if (b->name[lclen/8] & (1<<(lclen%8))) {
				b1->child[0] = b2;
				b1->child[1] = b;
			} else {
				b1->child[0] = b;
				b1->child[1] = b2;
			}
			return NULL;
		}
		if (len == b->len) {
			oldvalue = b->value;
			b->value = value;
			return oldvalue;
		}
		prevlen = b->len;
		if (name[prevlen/8] & (1<<(prevlen%8)))
			bp = &b->child[1];
		else
			bp = &b->child[0];
	}
}

#ifdef TEST

/*------------------------------------------------------------
 * test harness
 */

#include <string.h>

static void
print_name(name, len)
	unsigned char *name;
	int len;
{
	int i;

	for (i = 0; i < len; i++)
	    putchar((name[i/8] & (1<<(i%8))) ? '1' : '0');
}

static void
dump_tree(b, indent, leader)
	struct btree *b;
	int indent;
	char* leader;
{
	printf("%*s%s", indent, "", leader);
	if (b) {
		print_name(b->name, b->len);
		printf("/%d", b->len);
		if (b->value)
			printf(" = '%s'", (char *)b->value);
		putchar('\n');
		dump_tree(b->child[0], indent + 2, "0: ");
		dump_tree(b->child[1], indent + 2, "1: ");
	} else
		printf("-\n");
}

static void
apply_free(name, len, value, d)
	unsigned char *name;
	int len;
	void *value;
	void *d;
{
	free(value);
}

static void
apply_print(name, len, value, d)
	unsigned char *name;
	int len;
	void *value;
	void *d;
{
	print_name(name, len);
	printf(" -> '%s'\n", (char *)value);
}

int
main()
{
	struct btree *b = NULL;
	char buf[1024], cmd, *s;
	unsigned char name[64];
	void *value;
	int len = 0;

	for (;;) {
		printf("> ");
		s = gets(buf);
		if (!s)
			break;
		cmd = *s;
		if (*s)
			s++;
		while (*s == ' ' || *s == '\t')
			s++;
		if (cmd == '?') {
			if (*s) { printf("too many args\n"); continue; }
			printf("?                - help\n");
			printf("f <string>       - find a bitstring\n");
			printf("i <string> <val> - insert a bitstring\n");
			printf("d                - dump binary tree\n");
			printf("l                - list nodes in tree\n");
			printf("z                - reset to empty tree\n");
			printf("q                - quit\n");
			continue;
		}
		if (cmd == 'z') {
			if (*s) { printf("too many args\n"); continue; }
			btree_apply(b, apply_free, NULL);
			btree_free(b);
			b = NULL;
			continue;
		}
		if (cmd == 'd') {
			if (*s) { printf("too many args\n"); continue; }
			dump_tree(b, 0, "*: ");
			continue;
		}
		if (cmd == 'l') {
			if (*s) { printf("too many args\n"); continue; }
			btree_apply(b, apply_print, NULL);
			continue;
		}
		if (cmd == 'f' || cmd == 'i') {
			len = 0;
			if (*s == '-') {
			    /* empty name */
			    s++;
			} else
			    while (*s == '0' || *s == '1') {
				unsigned char bit = 1<<(len%8);
				name[len/8] &= ~bit;
				if (*s == '1') name[len/8] |= bit;
				s++;
				len++;
			    }
			while (*s == ' ' || *s == '\t')
				s++;

			if (cmd == 'f') {
				if (*s) { 
					printf("too many args\n"); 
					continue;
				}
				printf("find ");
				print_name(name, len);
				printf(" -> ");
				fflush(stdout);
				value = btree_find(b, name, len);
				if (value)
					printf("'%s'\n", (char *)value);
				else
					printf("(null)\n");
				continue;
			}

			else /* if (cmd == 'i') */ {
				value = *s ? strdup(s) : NULL;
				printf("insert ");
				print_name(name, len);
				if (value)
					printf(" '%s' -> ", (char *)value);
				else
					printf(" (null) -> ");
				fflush(stdout);
				value = btree_insert(&b, name, len, value);
				if (value) {
					printf("'%s'\n", (char *)value);
					free(value);
				} else
					printf("(null)\n");
				continue;
			}
		}

		if (cmd == 'q')
			break;

		printf("?\n");
	}
	exit(0);
}
#endif
