/*
 * Poor Man's Token Bus daemon
 */
/* David Leonard, 2003. Public domain. */

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <ifaddrs.h>
#include <netdb.h>
#include <poll.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <machine/endian.h>

#include <net/if.h>
#include <net/if_tun.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip6.h>

#include <arpa/inet.h>

#include "pmtb.h"
#include "btree.h"

static const char rcs[] = "$Id: pmtb.c,v 1.3 2003/07/27 02:56:56 d Exp $";
static const char version[] = "1.0";

/*------------------------------------------------------------
 * Types
 */

typedef u_int8_t byte;

#define MAX_NODE	256
#define INVALID_NODE	(-1)
#define MAX_ADDR	16

struct node {
	int valid;
	int next_node;
	int naddr;
	struct sockaddr bus_addr;
	struct sockaddr addr[MAX_ADDR];
};

struct packet {
	struct packet *next;
	int dstnode;
	int len;
	byte data[1];
};

enum state {
	S_INITIAL = 0,
	S_LEARNING,
	S_ATTACHING,
	S_HASTOKEN,
	S_SINGLE,
	S_IDLE,
	S_EXIT
};

/*------------------------------------------------------------
 * Prototypes
 */
void		siginfo(int sig);
void		sighup(int sig);
void		sigint(int sig);
void		sigalrm(int sig);

void		start_timer(int msec);
void		stop_timer(void);

void		broadcast(const byte *buf, int len);
void		broadcast1(int sym);
void		unicast(struct sockaddr *sa, const byte *buf, int len);
void		unicast1(struct sockaddr *sa, int sym);

const char *	addr_to_str(struct sockaddr *sa);
void		remove_addrs(struct node *n);
void		insert_addrs(struct node *n);
void		local_addr(struct node *n);
int		make_name(unsigned char *name, struct sockaddr *sa);

void		enqueue_data(const byte *buf, int len);
struct packet *	enqueue(int len);
struct packet *	dequeue(void);
void		forget(int id);

const char *	state_name(enum state state);
const char *	packet_name(const byte sym);
void		dump_info(void);
void		dump_buf(const byte *buf, int len);

void		announce(void);
int		learn(struct sockaddr *sa, const byte *buf, int len);

enum state	error(const char *fmt);
enum state	ignore(enum state state, const byte * buf, int len,
			struct sockaddr *from);
enum state	reset(void);
enum state	single(void);
enum state	attach(void);
enum state	have_token(void);
enum state	machine(enum state state, struct sockaddr *from, 
			unsigned char *buf, int len);

/*------------------------------------------------------------
 * Global state
 */

/*
 * List of nodes that we know are in the ring,
 * and a binary tree for reveres lookup by address.
 */
int nnode = 0;
struct node node[MAX_NODE];
struct btree *node_addr = NULL;

/*
 * Queue/buffer for the packets we send
 */
struct packet *head = NULL, *tail = NULL;

/*
 * Timer/timeout state
 */
volatile int timer_running = 0;
volatile int timer_expired = 0;

/*
 * Local sockets, addresses and devices
 */
const char *tunif, *busif;
struct sockaddr bus_bcast, bus_local;
int tunfd, bcast_sock, local_sock;

/*
 * This node's current state
 */
enum state state;
int this_node = INVALID_NODE;
int data_quantum = 32768;		/* max tx data per token */
unsigned long cycles = 0;		/* number of times token has cycled */
int want_announce = 0;			/* we must send ANNOUNCE asap */
struct sockaddr want_attach = { 0 };	/* node that last sent ATTACH */
volatile int want_quit = 0, 		/* SIGINT */
	     want_info = 0;		/* SIGINFO */

/*
 * Debugging
 */
#define dprintf(fmt, arg...) do { \
	if (dflag) fprintf(stderr, fmt , ## arg); } while (0)
int dflag = 0;

/*------------------------------------------------------------
 * Signal handlers
 */
void
siginfo(sig)
	int sig;
{
	want_info = 1;
}

void 
sighup(sig)
	int sig;
{
	want_announce = 1;
}

void 
sigint(sig)
	int sig;
{
	want_quit = 1;
}

void 
sigalrm(sig)
	int sig;
{
	if (timer_running) {
		timer_expired = 1;
		timer_running = 0;
	}
}

/*------------------------------------------------------------
 * Timers
 */

/*
 * Start a timer going. 
 *
 * The variable 'timer_expired' is set on expiry.
 * SIGALRM is generated on expiry, calling sigalrm(),
 * interrupting the poll() system call and setting
 * 'timer_expired' to 1.
 */
void
start_timer(msec)
	int msec;
{
	struct itimerval val;

	dprintf("[timer %s %d msec]\n", msec ? "start" : "stop", msec);

	val.it_value.tv_sec = msec / 1000;
	val.it_value.tv_usec = (msec % 1000) * 1000;
	timerclear(&val.it_interval);
	if (setitimer(ITIMER_REAL, &val, NULL) < 0)
		err(1, "setitimer");
	timer_running = msec ? 1 : 0;
	timer_expired = 0;
}

/*
 * Stop the timer.
 */
void
stop_timer()
{
	start_timer(0);
}

/*------------------------------------------------------------
 * Packet sending
 */

/*
 * Send a packet to the virtual ring broadcast address
 */
void
broadcast(buf, len)
	const byte *buf;
	int len;
{
	dprintf("broadcast %s/%d\n", packet_name(buf[0]), len);
	if (sendto(local_sock, buf, len, 0, &bus_bcast, bus_bcast.sa_len) < 0)
		err(1, "broadcast sendto %s", addr_to_str(&bus_bcast));
}

/*
 * Send a 1-byte packet to the virtual ring broadcast address
 */
void
broadcast1(sym)
	byte sym;
{
	broadcast(&sym, 1);
}

/*
 * Send a packet to the given node (unicast)
 */
void
unicast(sa, buf, len)
	struct sockaddr *sa;
	const byte *buf;
	int len;
{
	dprintf("unicast %s/%d to %s\n", packet_name(buf[0]), len, addr_to_str(sa));
	if (sendto(local_sock, buf, len, 0, sa, sa->sa_len) < 0)
		err(1, "unicast sendto %s", addr_to_str(sa));
}

/*
 * Send a 1-byte packet to the given node (unicast)
 */
void
unicast1(sa, sym)
	struct sockaddr *sa;
	byte sym;
{
	unicast(sa, &sym, 1);
}

/*------------------------------------------------------------
 * Address utilities 
 */

/*
 * Convert a sockaddr into a readable string.
 */
const char *
addr_to_str(sa)
	struct sockaddr *sa;
{
	static char buf[1024];
	char tmp[1024];
	const char *s = NULL;

	if (sa == NULL)
		return "(null)";
	if (sa->sa_len == 0)
		return "(zero)";

	switch (sa->sa_family) {
	case AF_INET:
	    {
		struct sockaddr_in *sin = (struct sockaddr_in *)sa;
		s = inet_ntoa(sin->sin_addr);
		snprintf(buf, sizeof buf, "%s:%u", s, ntohs(sin->sin_port));
		break;
	    }
	case AF_INET6:
	    {
		struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
		s = inet_ntop(AF_INET6, &sin6->sin6_addr, tmp, sizeof tmp);
		snprintf(buf, sizeof buf, "%s:%u", s, ntohs(sin6->sin6_port));
		break;
	    }
	default:
		snprintf(buf, sizeof buf, "<address family %d>",
		    sa->sa_family);
	}
	return buf;
}

/*
 * Remove all the addresses of a node from the btree
 */
void
remove_addrs(n)
	struct node *n;
{
	int i, namelen;
	unsigned char name[64];

	for (i = 0; i < n->naddr; i++) {
		namelen = make_name(name, &n->addr[i]);
		btree_insert(&node_addr, name, namelen, NULL);
	}
}

/*
 * Insert all the addresses of a node into the btree
 */
void
insert_addrs(n)
	struct node *n;
{
	int i, namelen;
	unsigned char name[64];

	for (i = 0; i < n->naddr; i++) {
		namelen = make_name(name, &n->addr[i]);
		btree_insert(&node_addr, name, namelen, n);
	}
}

/*
 * Set up our local addresses by reading addresses from
 * the tunnel interface.
 */
void
local_addr(n)
	struct node *n;
{
	struct ifaddrs *ifa_head, *ifa;
	struct sockaddr *sa;
	int found = 0;

	remove_addrs(n);

	if (getifaddrs(&ifa_head) < 0)
	    err(1, "getifaddrs");
	n->naddr = 0;
	for (ifa = ifa_head; ifa; ifa = ifa->ifa_next)
	    if (strcmp(ifa->ifa_name, tunif) == 0) {
		found = 1;

		/* Ignore unknown address families */
		if (ifa->ifa_addr->sa_family != AF_INET &&
		    ifa->ifa_addr->sa_family != AF_INET6)
			continue;

		/* Add this address to the list of local ring addresses */
		if (n->naddr < MAX_ADDR) {
			sa = &n->addr[n->naddr++];
			memcpy(sa, ifa->ifa_addr, sizeof *sa);
			dprintf("\t+ %s\n", addr_to_str(sa));
		} else {
			warn("%s: too many addresses", tunif);
			break;
		}
	    }
	freeifaddrs(ifa_head);
	if (!n->naddr) {
	    if (!found)
	    	errx(1, "%s: interface not found", tunif);
	    else
	    	errx(1, "%s: interface has no active addresses", tunif);
	}

	insert_addrs(n);
}

/*------------------------------------------------------------
 * Data packet utilties
 */

/*
 * Create a name for the btree based on the given sockaddr.
 */
int
make_name(name, sa)
	unsigned char *name;
	struct sockaddr *sa;
{
	static unsigned char reverse[256];
	static int reverse_generated = 0;
	unsigned char *s;
	int namelen, i;

	/*
	 * Dynamic construction of bit reversing table
	 */
	if (!reverse_generated) {
		static unsigned char r4[16] = {
			0, 8, 4, 12, 2, 10, 6, 14,
			1, 9, 5, 13, 3, 11, 7, 15 };
		for (i = 0; i < 256; i++) 
			reverse[i] = (r4[i & 15] << 4) |
			   	      r4[(i >> 4) & 15];
		reverse_generated = 0;
	}

	switch (sa->sa_family) {
	case AF_INET:
		namelen = sizeof (struct in_addr);
		s = (char *)&((struct sockaddr_in *)sa)->sin_addr;
		for (i = 0; i < namelen; i++)
			name[namelen - i] = reverse[s[i]];
		return namelen;
	case AF_INET6:
		namelen = sizeof (struct in6_addr);
		s = (char *)&((struct sockaddr_in6 *)sa)->sin6_addr;
		for (i = 0; i < namelen; i++)
			name[namelen - i] = reverse[s[i]];
		return namelen;

	default:
		err(1, "make_name: family not implemented");
	}
}

/* 
 * A packet is being sent into the tun device. 
 * We figure out which node it is to go to and
 * add it to the send queue for later.
 */
void
enqueue_data(buf, len)
	const byte *buf;
	int len;
{
	struct packet *pkt;
	struct sockaddr sa;
	unsigned char name[64];
	int namelen;
	struct node *n;
	u_int32_t af;

	/* Find the destination address in the packet header */
	memcpy(&af, buf, sizeof af);	/* in case of alignment */
	af = ntohl(af);

	/* Create a sockaddr representing the target host */
	switch (af) {
	case AF_INET:
	    {
		struct ip *ip = (struct ip *)(buf + sizeof af);
		memset(&sa, 0, sizeof sa);
		sa.sa_len = sizeof (struct sockaddr_in);
		sa.sa_family = AF_INET;
		memcpy(&((struct sockaddr_in *)&sa)->sin_addr,
			&ip->ip_dst, sizeof (struct in_addr));
		break;
	    }
	case AF_INET6:
	    {
		struct ip6_hdr *ip6 = (struct ip6_hdr *)(buf + sizeof af);
		memset(&sa, 0, sizeof sa);
		sa.sa_len = sizeof (struct sockaddr_in6);
		sa.sa_family = AF_INET6;
		memcpy(&((struct sockaddr_in6 *)&sa)->sin6_addr,
			&ip6->ip6_dst, sizeof (struct in6_addr));
		break;
	    }
	default:
		/* silently drop unknown address families */
		dprintf("enqueue_data(): dropped bad family\n");
		return;
	}

	/*
	 * XXX should apply system routes here
	 */
	namelen = make_name(name, &sa);
	n = (struct node *)btree_find(node_addr, name, namelen);
	if (!n) {
#if 0
		/* Generate a No Route To Host packet */
		if (af == AF_INET) {
		    struct ip *ip = (struct ip *)(buf + sizeof af);
		    struct {
		       u_int32_t af;
		       struct { 
			struct ip ip;
			struct icmp icmp;
		       } p;
		    } pkt;

		    memset(&pkt, 0, sizeof pkt);
		    pkt.af = htonl(af);
		    pkt.p.ip.ip_v = IPVERSION;
		    pkt.p.ip.ip_hl = 0;
		    pkt.p.ip.ip_len = sizeof pkt;
		    pkt.p.ip.ip_ttl = 255;
		    pkt.p.ip.ip_p = IPPROTO_ICMP;
		    memcpy(&pkt.p.ip.ip_dst, &ip->ip_src, 
			sizeof (struct in_addr));

		    pkt.p.icmp.icmp_type = ICMP_UNREACH;
		    pkt.p.icmp.icmp_code = ICMP_UNREACH_HOST;
		    pkt.p.icmp.icmp_cksum = 0;

		    write(tunfd, pkt, sizeof pkt);
		}
#endif
		return;	
	}

	if (n < node || n >= &node[MAX_NODE] || !n->valid)
		errx(1, "enqueue_data()");

	pkt = enqueue(len + 4);
	if (pkt) {
		pkt->dstnode = n - node;
		pkt->data[0] = P_DATA;
		pkt->data[1] = this_node;
		pkt->data[2] = pkt->dstnode;
		pkt->data[3] = 0;
		memcpy(pkt->data + 4, buf, len);
		dprintf("enqueue_data: %d bytes queued for node %02x", 
		    len, pkt->dstnode);
	} else {
		dprintf("enqueue_data: no memory!");
	}
}

/*
 * Create & enqueue an empty unicast packet.
 * Caller should fill pkt->data and pkt->dstnode.
 */
struct packet *
enqueue(len)
	int len;
{
	struct packet *pkt;

	pkt = (struct packet *)malloc(sizeof (struct packet) + len - 1);
	if (pkt == NULL)
		return NULL;

	pkt->dstnode = INVALID_NODE;
	pkt->len = len;

	/* Add to the end of the queue */
	if (tail)
		tail->next = pkt;
	else
		head = pkt;
	tail = pkt;
	pkt->next = NULL;

	return pkt;
}

/* 
 * Remove a packet from the send queue.
 * Caller should free() the packet on return.
 */
struct packet *
dequeue()
{
	struct packet *pkt;

	pkt = head;
	if (pkt) {
		head = pkt->next;
		if (!head)
		    tail = NULL;
	}
	return pkt;
}

/*
 * Remove packets in the queue destined for the
 * given node id
 */
void
forget(id)
	int id;
{
	struct packet *p, **pktp;

	tail = NULL;
	for (pktp = &head; *pktp; pktp = &(tail = *pktp)->next)
	    while (*pktp && (*pktp)->dstnode == id) {
		p = *pktp;
		*pktp = p->next;
		free(p);
	    }
}

/*------------------------------------------------------------
 * Debugging
 */

/*
 * Return the name for a state
 */
const char *
state_name(state)
	enum state state;
{
	switch (state) {
	case S_INITIAL:		return "S_INITIAL";
	case S_LEARNING:	return "S_LEARNING";
	case S_ATTACHING:	return "S_ATTACHING";
	case S_HASTOKEN:	return "S_HASTOKEN";
	case S_SINGLE:		return "S_SINGLE";
	case S_IDLE:		return "S_IDLE";
	case S_EXIT:		return "S_EXIT";
	default:		return "<*bad state*>";
	}
}

/*
 * Return the name for a packet type
 */
const char *
packet_name(sym)
	const byte sym;
{
	switch (sym) {
	case P_RESET:		return "P_RESET";
	case P_TOKEN:		return "P_TOKEN";
	case P_DATA:		return "P_DATA";
	case P_LEARN:		return "P_LEARN";
	case P_ATTACH:		return "P_ATTACH";
	case P_BUSY:		return "P_BUSY";
	case P_ANNOUNCE:	return "P_ANNOUNCE";
	case P_LEAVE:		return "P_LEAVE";
	case P_TIMEOUT:		return "P_TIMEOUT";
	default: 
	    {
		static char desc[16];
		snprintf(desc, sizeof desc, "<0x%02x>", sym);
		return desc;
	    }
	}
}

/*
 * Debugging: show current node state
 */
void
dump_info()
{
	int i, j;
	struct packet *pkt;
	int nodecount = 0;
	struct node *n;

	fprintf(stderr, "state = %s\n", state_name(state));
	fprintf(stderr, "this_node = %02x\n", this_node);
	fprintf(stderr, "cycles = %ld\n", cycles);

	fprintf(stderr, "bus interface %s at %s ", busif, 
		addr_to_str((struct sockaddr *)&bus_local));
	fprintf(stderr, "broadcast %s\n", 
		addr_to_str((struct sockaddr *)&bus_bcast));

	fprintf(stderr, "nnode=%d:\n", nnode);
	for (i = 0; i < MAX_NODE; i++)
	    if (node[i].valid) {
		n = &node[i];
		nodecount++;
		fprintf(stderr, "  %c%02x:", (i == this_node) ? '*' : ' ', i);
		if (n->naddr == 0)
		    fprintf(stderr, "\t<no addresses>\n");
		else
		    for (j = 0; j < n->naddr; j++)
			switch (n->addr[j].sa_family) {
			case AF_INET:
				fprintf(stderr, "\tinet  %s\n", 
				    addr_to_str(&n->addr[j]));
				break;
			case AF_INET6:
				fprintf(stderr, "\tinet6 %s\n", 
				    addr_to_str(&n->addr[j]));
				break;
			default:
				fprintf(stderr, "\t<unknown family %d>\n",
				    n->addr[j].sa_family);
			}
		fprintf(stderr, "\tbus_addr = %s\n", addr_to_str(&n->bus_addr));
		fprintf(stderr, "\tnext_node = %02x\n", n->next_node);
	    }

	if (nodecount != nnode)
		fprintf(stderr, "*** nnode is wrong ***\n");

	i = 0;
	for (pkt = head; pkt; pkt = pkt->next)
		i++;
	fprintf(stderr, "queue: %d entr%s\n", i, i == 1 ? "y" : "ies");

	if (want_quit)
		printf("want_quit = 1\n");
	if (want_announce)
		printf("want_announce = 1\n");
	printf("want_attach = %s\n", 
	    addr_to_str(want_attach.sa_len ? &want_attach : NULL));

	fprintf(stderr, "\n");
}

/*
 * Debugging: dump contents of a packet to stderr.
 */
void
dump_buf(buf, len)
	const byte *buf;
	int len;
{
	int i, j;

	for (i = 0; i < len; i += 16) {
	    fprintf(stderr, "\t%04x:", i);
	    for (j = i; j < i + 16; j++)
		if (j < len)
		    fprintf(stderr, " %02x", buf[j]);
		else
		    fprintf(stderr, "   ");
	    fprintf(stderr, "\n");
	}
}

/*------------------------------------------------------------
 * State machine utilities
 */

/* 
 * Construct and send our ANNOUNCE packet, so that
 * other nodes know what we do and have.
 */
void
announce()
{
	static byte buf[1024];
	int i;
	int len;
	struct node *n;

	/* Re-load our tunnel addresses */
	n = &node[this_node];
	local_addr(n);

	buf[0] = P_ANNOUNCE;
	buf[1] = this_node;
	buf[2] = n->next_node;
	buf[3] = nnode;
	len = 4;
	for (i = 0; i < n->naddr; i++) {
		memcpy(&buf[len], &n->addr[i], n->addr[i].sa_len);
		len += n->addr[i].sa_len;
	}

	dprintf("announce():\n");
	if (dflag) dump_buf(buf, len);

	broadcast(buf, len);
}

/*
 * Called on an ANNOUNCE or LEAVE.
 * We update our node table based on the received packet.
 * Returns 0 on error, or 1 on success.
 */
int
learn(sa, buf, len)
	struct sockaddr *sa;
	const byte *buf;
	int len;
{
	struct node *n;
	int i, id;
	unsigned char name[64];
	int namelen;

	/* Remove a node from the list */
	if (buf[0] == P_LEAVE) {
	    id = buf[1];
	    n = &node[id];
	    if (!n->valid || (this_node != INVALID_NODE && id == this_node))
	    {
		dprintf("learn(LEAVE): node %02x not valid or self\n", id);
		return 0;
	    }
	    /* Maybe also check that (*sa == n->bus_addr) ?? */
	    n->valid = 0;
	    nnode--;

	    /* close chain if it was our 'next' node */
	    if (this_node != INVALID_NODE && 
		id == node[this_node].next_node) {
		node[this_node].next_node = n->next_node;
		want_announce = 1;
	    }

	    /* Remove stuff from the packet queue */
	    forget(id);

	    return 1;
	}

	/* otherwise it is an announce */
	if (buf[0] != P_ANNOUNCE)
		errx(1, "learn()");

	n = &node[buf[1]];
	if (!n->valid) {
		memset(n, 0, sizeof *n);
		n->valid = 1;
		nnode++;
	}
	if (this_node != INVALID_NODE && this_node == buf[1]) {
		dprintf("learn(ANNOUNCE): node %02x is us!\n", buf[1]);
		return 0;
	}

	/* if (buf[3] != nnode) return 0; */

	/* First, remove all of its old addresses */
	remove_addrs(n);

	n->naddr = 0;
	n->next_node = buf[2];
	memcpy(&n->bus_addr, sa, sizeof n->bus_addr);

	/* Carefully extract sockaddr structures from the packet */
	i = 4;
	while (i < len) {
		struct sockaddr *sa = (struct sockaddr *)(buf + i);
		if (sa->sa_len + i > len ||
		    sa->sa_len > sizeof (struct sockaddr))
		{
			dprintf("learn(ANNOUNCE): bad sa_len=%d\n",
				sa->sa_len);
			if (dflag) dump_buf(buf, len);
			return 0;
		}

		/* Add the address to the node's address list */
		memcpy(&n->addr[n->naddr], sa, sa->sa_len);
		n->naddr++;
		i += sa->sa_len;

		/* Insert the address into the btree for easy finding */
		namelen = make_name(name, sa);
		btree_insert(&node_addr, name, namelen, n);
	}
	if (i != len) {
		dprintf("learn(ANNOUNCE): short packet? i=%d len=%d\n", i, len);
		if (dflag) dump_buf(buf, len);
		return 0;
	}

	return 1;
}


/*------------------------------------------------------------
 * State machine reactions
 */

/*
 * Called when we get some kind of error.
 * We generate a RESET to bring the ring back to order.
 */
enum state
error(fmt)
	const char *fmt;
{
	warnx("%s", fmt);
	broadcast1(P_RESET);
	return reset();
}

/*
 * Called when a packet is ignored.
 * (Mostly useful for debugging)
 */
enum state
ignore(state, buf, len, from)
	enum state state;
	const byte *buf;
	int len;
	struct sockaddr *from;
{
	warnx("ignorning packet %s/%d in state %s from %s\n", 
		buf ? packet_name(*buf) : "(null)", 
		len, state_name(state),
		addr_to_str(from));
	return state;
}

/*
 * Clear all tables and enter the initial state.
 */
enum state
reset()
{
	struct packet *pkt;


	dprintf("reset()\n");
	stop_timer();

	/* Reset the node state */
        memset(node, 0, sizeof node);
	btree_free(node_addr);
	node_addr = NULL;
	nnode = 0;
	cycles = 0;

	/* Empty the packet queue! */
	while ((pkt = dequeue()) != NULL)
		free(pkt);

	this_node = INVALID_NODE;
	want_announce = 0;
	want_attach.sa_len = 0;

	start_timer(1000 + random() % 1000);

	if (want_quit)
		exit(0);

	return S_INITIAL;
}

/*
 * Called when we determine that we have to become an isolated
 * node. We reset, become node #0 and enter S_SINGLE.
 */
enum state
single()
{
	dprintf("** single\n");

	broadcast1(P_RESET);	/* generate reset */
	reset();
	stop_timer();

	/* Become node # 0 */
	this_node = 0;
	node[this_node].valid = 1;
	nnode = 1;
	local_addr(&node[this_node]);

	return S_SINGLE;
}

/* 
 * Called to send an ATTACH to a random node and enter S_ATTACHING.
 */
enum state
attach()
{
	int i;

	if (nnode == 0)
		errx(1, "attach()");

	/* Find a random node */
	i = random() % MAX_NODE;
	while (!node[i].valid) 
		i = (i + 1) % MAX_NODE;

	unicast1(&node[i].bus_addr, P_ATTACH);

	start_timer(1000);
	return S_ATTACHING;
}

/*
 * Send pending data, annoucements, etc.
 * i.e. all the things we do when we have the token.
 */
enum state
have_token()
{
	int i;
	struct packet *pkt;
	int sent_data;
	byte token[4];

	cycles++;

	/*
	 * If someone wanted to attach, we attach them here now
	 */
	if (want_attach.sa_len) {
	    if (want_quit) 
		unicast1(&want_attach, P_BUSY);
	    else {
		/* Find a free node id */
		for (i = 0; i < MAX_NODE; i++)
		    if (!node[i].valid)
		    {
			/* Fill this free slot with basic info */
			memset(&node[i], 0, sizeof node[i]);
			memcpy(&node[i].bus_addr, &want_attach, sizeof
			       node[i].bus_addr);
			node[i].valid = 1;
			nnode++;

			/* Insert them into the ring ahead of us */
			node[i].next_node = node[this_node].next_node;
			node[this_node].next_node = i;	
			want_announce = 1;
			break;
		    }
	    }
	    want_attach.sa_len = 0;
	}

	/*
	 * If we had announcements to send, send them now
	 */
	if (want_announce) {
		announce();
		want_announce = 0;
	}

	/*
	 * Send pending data, if any
	 */
	sent_data = 0;
	while (sent_data < data_quantum && (pkt = dequeue()) != NULL) {
	    if (node[pkt->dstnode].valid) {
		unicast(&node[pkt->dstnode].bus_addr, pkt->data, pkt->len);
		sent_data += pkt->len;
	    } 
	    free(pkt);
	}

	/*
	 * If we are quitting, send a LEAVE packet.
	 */
	if (want_quit) {
		byte leave[4];

		leave[0] = P_LEAVE;
		leave[1] = this_node;
		leave[2] = node[this_node].next_node;
		leave[3] = nnode - 1;
		broadcast(leave, sizeof leave);
	}

	/*
	 * Now that we're done, we figure out where to 
	 * send the token on to.
	 */
	if (nnode == 1)
		return want_quit ? S_EXIT : S_SINGLE;
	else {
		byte next_node = node[this_node].next_node;

		token[0] = P_TOKEN;
		token[1] = this_node;
		token[2] = next_node;
		token[3] = node[next_node].next_node;
		unicast(&node[next_node].bus_addr, token, sizeof token);
		return want_quit ? S_EXIT : S_IDLE;
	}
}

/*
 * Master protocol state machine
 */
enum state
machine(state, from, buf, len)
	enum state state;
	struct sockaddr *from;
	unsigned char *buf;
	int len;
{
    byte sym = buf[0];

    dprintf("machine(%s, %s, %s, %d)\n",
	state_name(state), addr_to_str(from),
	packet_name(sym), len);

    switch (state) {

    case S_INITIAL:
	if (want_quit)
		return S_EXIT;
	switch (sym) {
	case P_TIMEOUT:
		broadcast1(P_LEARN);
		start_timer(1000);
		return S_LEARNING;
	case P_RESET:
		return reset();
	default:
		return S_INITIAL;
	}

    case S_LEARNING:
	if (want_quit)
		return S_EXIT;
	switch (sym) {
	case P_TIMEOUT:
		if (nnode == 0)
			return single();
		return error("timed out learning ring");
	case P_RESET:
		return reset();
	case P_LEAVE:
		if (node[buf[1]].valid && !learn(from, buf, len))
			return error("learnt bad leave");
		if (nnode == buf[3]) 
			return attach();
		return S_LEARNING;
	case P_ANNOUNCE:
		if (!learn(from, buf, len))
			return error("learnt bad announce");
		if (nnode == buf[3])
			return attach();
		return S_LEARNING;
	case P_LEARN:
		return S_LEARNING;
	default:
		return ignore(state, buf, len, from);
	}

    case S_ATTACHING:
	if (want_quit) {
		broadcast1(P_RESET);
		return S_EXIT;
	}
	switch (sym) {
	case P_TIMEOUT:
		return error("timed out attaching");
	case P_BUSY:
		return attach();	/* try another node */
	case P_LEAVE:
	case P_ANNOUNCE:
		if (!learn(from, buf, len))
			return error("bad announce/leave while attaching");
		return S_ATTACHING;
	case P_TOKEN:
		if (node[buf[2]].valid)
			return error("attached as node already in use");
		this_node = buf[2];
		memset(&node[this_node], 0, sizeof node[this_node]);
		node[this_node].next_node = buf[3];
		node[this_node].valid = 1;
		nnode++;
		want_announce = 1;
		return have_token();
	case P_RESET:
		return reset();
	default:
		return ignore(state, buf, len, from);
	}

    case S_SINGLE:
	if (want_quit) {
		broadcast1(P_RESET);
		return S_EXIT;
	}
	switch (sym) {
	case P_LEARN:
		announce();
		return S_SINGLE;
	case P_ATTACH:
		memcpy(&want_attach, from, sizeof want_attach);
		return have_token();
	case P_TIMEOUT:
		/* ignore */
		return S_SINGLE;
	case P_ANNOUNCE:
	case P_LEAVE:
		/* ignore - probably a leftover */
		return S_SINGLE;
	case P_RESET:
		return reset();
	default:
		return ignore(state, buf, len, from);
	}

    case S_IDLE:
	switch (sym) {
	case P_TIMEOUT:
		return error("timed out idling for token");
	case P_ANNOUNCE:
	case P_LEAVE:
		if (!learn(from, buf, len))
			return error("bad announce/leave");
		return S_IDLE;
	case P_LEARN:
		want_announce = 1;
		return S_IDLE;
	case P_ATTACH:
		if (want_attach.sa_len)
			/* XXX should be queued for sending when we have token */
			unicast1(from, P_BUSY);
		else
			memcpy(&want_attach, from, sizeof want_attach);
		return S_IDLE;
	case P_TOKEN:
		if (buf[2] != this_node)
			return error("received someone else's token");
		if (buf[3] != node[this_node].next_node)
			want_announce = 1;
		start_timer(1000);
		return have_token();
	case P_DATA:
		if (buf[2] != this_node)
			return error("receieved someone else's data");
		if (write(tunfd, buf + 4, len - 4) < 0)
			warn("%s: write", tunif);
		return S_IDLE;
	case P_RESET:
		return reset();
	default:
		return ignore(state, buf, len, from);
	}

    default:
	errx(1, "machine(): invalid state");
    }
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct ifaddrs *ifa_head, *ifa;
	struct sockaddr from;
	socklen_t from_len;
	const char *s;
	char tundev[PATH_MAX];
	byte buf[16384];
	int i, len, ch, argerr = 0;
	struct pollfd pfd[3];

	srandomdev();

	/*
	 * Process command line arguments
	 */
	while ((ch = getopt(argc, argv, "dv")) != -1)
	    switch (ch) {
	    case 'd':
		dflag++;
		break;
	    case 'v':
		fprintf(stderr, "pmtb version %s\n", version);
		exit(0);
	    default:
		argerr = 1;
	    }
	if (optind + 2 != argc)
		argerr = 1;
	if (argerr) {
		fprintf(stderr, "usage: %s [-dv] tunif busif\n", 
		    argv[0]);
		exit(1);
	}

	tunif = argv[optind++];
	busif = argv[optind++];

	/*
	 * Open the tunnel device
	 */
	if (memcmp(tunif, "tun", 3) != 0 || tunif[3] == '\0')
	    errx(1, "%s: bad tunnel device name", tunif);
	for (s = tunif + 3; *s; s++)
	    if (*s < '0' || *s > '9')
	        errx(1, "%s: bad tunnel device name", tunif);
	snprintf(tundev, sizeof tundev, "/dev/%s", tunif);
	if ((tunfd = open(tundev, O_RDWR, 0666)) < 0)
		err(1, "%s", tundev);

	/*
	 * Bring the tunnel interface up, in broadcast mode
	 */
	i = IFF_UP | IFF_BROADCAST;
	if (ioctl(tunfd, TUNSIFMODE, &i) < 0)
		err(1, "%s TUNSIFMODE", tundev);

	/* 
	 * Look for the bus interface
	 */
	if (getifaddrs(&ifa_head) < 0)
	    err(1, "getifaddrs");
	for (ifa = ifa_head; ifa; ifa = ifa->ifa_next)
	    if (strcmp(ifa->ifa_name, busif) == 0 &&
		ifa->ifa_addr->sa_family == AF_INET &&
		(ifa->ifa_flags & (IFF_UP|IFF_BROADCAST)) == 
			(IFF_UP|IFF_BROADCAST))
		{
		    memcpy(&bus_bcast, ifa->ifa_broadaddr, sizeof bus_bcast);
		    memcpy(&bus_local, ifa->ifa_addr, sizeof bus_local);
		    break;
	    	}
	freeifaddrs(ifa_head);
	if (!ifa)
	    errx(1, "%s: no AF_INET broadcast links configured", busif);

	dprintf("bus interface %s at %s ", busif, 
		addr_to_str((struct sockaddr *)&bus_local));
	dprintf("broadcast %s\n", 
		addr_to_str((struct sockaddr *)&bus_bcast));

	/*
	 * Construct a socket for hearing broadcast UDP
	 */
	if ((bcast_sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
		err(1, "bcast socket");
	i = 1;
	if (setsockopt(bcast_sock, SOL_SOCKET, SO_REUSEPORT, &i, sizeof i) < 0)
		err(1, "bcast so_reuseport");
	if (setsockopt(bcast_sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof i) < 0)
		err(1, "bcast so_reuseaddr");
	((struct sockaddr_in *)&bus_bcast)->sin_port = htons(CONTROL_PORT);
	if (bind(bcast_sock, &bus_bcast, sizeof bus_bcast) < 0)
		err(1, "bcast bind");
	dprintf("bound bcast addr %s\n", addr_to_str(&bus_bcast));

	/*
	 * Construct a socket for hearing unicast UDP
	 */
	if ((local_sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
		err(1, "local socket");
	((struct sockaddr_in *)&bus_local)->sin_port = htons(0);
	if (bind(local_sock, &bus_local, sizeof bus_local) < 0)
		err(1, "local bind");
	i = sizeof bus_local;
	if (getsockname(local_sock, &bus_local, &i) < 0)
		err(1, "local getsockname");
	dprintf("bound local addr %s\n", addr_to_str(&bus_local));
	i = 1;
	if (setsockopt(local_sock, SOL_SOCKET, SO_BROADCAST, &i, sizeof i) < 0)
		err(1, "local so_broadcast");
	if (setsockopt(local_sock, SOL_SOCKET, SO_DONTROUTE, &i, sizeof i) < 0)
		err(1, "local so_dontroute");

	/*
	 * Set up signal handlers
	 */
	if (signal(SIGINT, sigint) == SIG_ERR)
		err(1, "signal SIGINT");
	if (signal(SIGTERM, sigint) == SIG_ERR)
		err(1, "signal SIGTERM");
	if (signal(SIGINFO, siginfo) == SIG_ERR)
		err(1, "signal SIGINFO");
	if (signal(SIGHUP, sighup) == SIG_ERR)
		err(1, "signal SIGHUP");
	if (signal(SIGALRM, sigalrm) == SIG_ERR)
		err(1, "signal SIGALRM");

	state = reset();

	while (state != S_EXIT) {

		if (want_info) {
			want_info = 0;
			dump_info();
		}

		if (want_quit && !timer_running) {
			/*
			 * XXX no timer means no chance of timing out,
			 * so, we just crash.
			 */
			break;
		}

		if (dflag) {
		    static enum state old_state = S_EXIT;
		    if (old_state != state) { 
		        old_state = state;
		        dprintf("(state now %s)\n", state_name(state));
		    }
		}

		if (timer_expired) {
			timer_expired = 0;
			buf[0] = P_TIMEOUT;
			state = machine(state, NULL, buf, 1);
			continue;
		}

		/*
		 * Wait for an event
		 */
		pfd[0].fd = bcast_sock;
		pfd[0].events = POLLIN;
		pfd[0].revents = 0;
		pfd[1].fd = local_sock;
		pfd[1].events = POLLIN;
		pfd[1].revents = 0;
		pfd[2].fd = tunfd;
		pfd[2].events = POLLIN;
		pfd[2].revents = 0;
		if (poll(pfd, 2, INFTIM) < 0)
		{
			if (errno == EINTR)
				continue;
			err(1, "poll");
		}

		/*
		 * Handle broadcast packets heard
		 */
		if (pfd[0].revents & POLLIN) {
		    from_len = sizeof from;
		    len = recvfrom(pfd[0].fd, buf, sizeof buf, 0,
			    &from, &from_len);
		    from.sa_len = from_len;
		    if (len < 0)
			    err(1, "bcast recvfrom");
		    if (len == 0)
			    errx(1, "bcast recvfrom: zero");
		    if (memcmp(&from, &bus_local, bus_local.sa_len) != 0) {
			switch (buf[0]) {
			case P_RESET:
				if (len != 1) goto ignore_bcast;
				break;
			case P_LEARN:
				if (len != 1) goto ignore_bcast;
				break;
			case P_ANNOUNCE:
				if (len < 4) goto ignore_bcast;
				break;
			case P_LEAVE:
				if (len != 4) goto ignore_bcast;
				break;
			default:
			ignore_bcast:
			        state = ignore(state, buf, len, &from);
				continue;
			}
			state = machine(state, &from, buf, len);
		    }
		}

		/*
		 * Handle unicast packets heard
		 */
		if (pfd[1].revents & POLLIN) {
		    from_len = sizeof from;
		    len = recvfrom(pfd[1].fd, buf, sizeof buf, 0,
			    &from, &from_len);
		    from.sa_len = from_len;
		    if (len < 0)
			    err(1, "local recvfrom");
		    if (len == 0)
			    errx(1, "local recvfrom: zero");
		    switch (buf[0]) {
		    case P_TOKEN:
			    if (len != 4) goto ignore_unicast;
			    break;
		    case P_DATA:
			    if (len < 8) goto ignore_unicast;
			    break;
		    case P_ATTACH:
			    if (len != 1) goto ignore_unicast;
			    break;
		    case P_BUSY:
			    if (len != 1) goto ignore_unicast;
			    break;
		    default:
		    ignore_unicast:
			    state = ignore(state, buf, len, &from);
			    continue;
		    }
		    state = machine(state, &from, buf, len);
		}

		/*
		 * Handle packets sent into the tun device
		 */
		if (pfd[2].revents & POLLIN) {
			len = read(pfd[2].fd, buf, sizeof buf);
			if (len < 0)
				err(1, "%s: read", tundev);
			if (len == 0)
				errx(1, "%s: read zero", tundev);
			enqueue_data(buf, len);
		}
	}

	dprintf("exiting with state = %s\n", state_name(state));

	exit(0);
}
